!**************************************************************************************************
! Module                      | decodtool_module
! Author                      | F. Guillaume (CNRM/GMAP/OBS)
! Creation Date               | 01/06/2016
!--------------------------------------------------------------------------------------------------
! Description : subroutines for checking arguments passed and printing sections of bufr files
!--------------------------------------------------------------------------------------------------
! Global Variables : 
!--------------------------------------------------------------------------------------------------
! History :
!           27/12/2016 - bugfix in PrintSec4()
!**************************************************************************************************

MODULE decodtool_module

use            :: parkind1,     only : jpim, jpib, jprb
use            :: bufr_module

implicit none

 character(len=1024) :: ErrString

contains


!**************************************************************************************************
! Subroutine/Function/Program | PrintSec0
! Author                      | F. Guillaume (CNRM/GMAP/OBS)
! Creation Date               | 01/06/2016
!--------------------------------------------------------------------------------------------------
! Description : print section 0 of bufr file if needed
!--------------------------------------------------------------------------------------------------
! Global Variables : 
!--------------------------------------------------------------------------------------------------
! History :
!**************************************************************************************************
subroutine PrintSec0(NumFic,Version,LPrtSec)
 integer(kind=jpim), intent(in)  :: NumFic
 logical,intent(in)              :: LPrtSec
 integer(kind=jpim), intent(out) :: Version

write(*,'(//,A," BUFR file n° ",I6.6,/)') repeat("*",11), NumFic
if (LPrtSec) then
  write(*,'("SECTION 0 : ",A)') repeat("-",54)
  write(*,'(TR12,"Total length of message (bytes)           = ",I10)') ksec0(2)
  write(*,'(TR12,"Length of section 0 (bytes)               = ",I10)') ksec0(1)
  write(*,'(TR12,"Bufr edition number                       = ",I10,/)') ksec0(3)
endif
version = ksec0(3)
end subroutine PrintSec0


!**************************************************************************************************
! Subroutine/Function/Program | PrintSec1
! Author                      | F. Guillaume (CNRM/GMAP/OBS)
! Creation Date               | 01/06/2016
!--------------------------------------------------------------------------------------------------
! Description : print section 1 of bufr file
!--------------------------------------------------------------------------------------------------
! Global Variables : 
!--------------------------------------------------------------------------------------------------
! History :
!**************************************************************************************************
subroutine PrintSec1(Version,LprtSec)
 integer(kind=jpim), intent(in) :: Version
 logical,intent(in)             :: LPrtSec

if (LPrtSec) then
  write(*,'("SECTION 1 : ",A)') repeat("-",54)
  select case (Version)
  case (3)
    write(*,'(TR12,"Length of section 1 (bytes)               = ",I10)') ksec1(1)
    write(*,'(TR12,"Bufr master table                         = ",I10)') ksec1(14)
    write(*,'(TR12,"Version number of master table            = ",I10)') ksec1(15)
    write(*,'(TR12,"Version number of local tables            = ",I10)') ksec1(8)
    write(*,'(TR12,"Originating centre (CCT C1)               = ",I10)') ksec1(3)
    write(*,'(TR12,"Originating sub-centre (CCT C12)          = ",I10)') ksec1(16)
    write(*,'(TR12,"Data category (table A)                   = ",I10)') ksec1(6)
    write(*,'(TR12,"Local Data sub-category                   = ",I10)') ksec1(7)
    write(*,'(TR12,"Update sequence number                    = ",I10)') ksec1(4)
    write(*,'(TR12,"Year                                      = ",I10)') ksec1(9)
    write(*,'(TR12,"Month                                     = ",I10)') ksec1(10)
    write(*,'(TR12,"Day                                       = ",I10)') ksec1(11)
    write(*,'(TR12,"Hour                                      = ",I10)') ksec1(12)
    write(*,'(TR12,"Minute                                    = ",I10)') ksec1(13)
    write(*,'(TR12,"Flag for presence of section 2            = ",I10,/)') ksec1(5)
  case (4)
    write(*,'(TR12,"Length of section 1 (bytes)               = ",I10)') ksec1(1)
    write(*,'(TR12,"Bufr master table                         = ",I10)') ksec1(14)
    write(*,'(TR12,"Version number of master table            = ",I10)') ksec1(15)
    write(*,'(TR12,"Version number of local tables            = ",I10)') ksec1(8)
    write(*,'(TR12,"Originating centre (CCT C11)              = ",I10)') ksec1(3)
    write(*,'(TR12,"Originating sub-centre (CCT C12)          = ",I10)') ksec1(16)
    write(*,'(TR12,"Data category (table A)                   = ",I10)') ksec1(6)
    write(*,'(TR12,"Local data sub-category                   = ",I10)') ksec1(7)
    write(*,'(TR12,"International data sub-category (CCT C13) = ",I10)') ksec1(17)
    write(*,'(TR12,"Update sequence number                    = ",I10)') ksec1(4)
    write(*,'(TR12,"Year                                      = ",I10)') ksec1(9)
    write(*,'(TR12,"Month                                     = ",I10)') ksec1(10)
    write(*,'(TR12,"Day                                       = ",I10)') ksec1(11)
    write(*,'(TR12,"Hour                                      = ",I10)') ksec1(12)
    write(*,'(TR12,"Minute                                    = ",I10)') ksec1(13)
    write(*,'(TR12,"Second                                    = ",I10)') ksec1(18)
    write(*,'(TR12,"Flag for presence of section 2            = ",I10,/)') ksec1(5)
  case (:2,5:)
    call erreur("***** ERROR : don't handle this version of Bufr edition")
  end select
endif
end subroutine PrintSec1


!**************************************************************************************************
! Subroutine/Function/Program | PrintSec3
! Author                      | F. Guillaume (CNRM/GMAP/OBS)
! Creation Date               | 01/06/2016
!--------------------------------------------------------------------------------------------------
! Description : print first part section 3 of bufr file, unexpanded sequence descriptor, in simple
!               or in param.cfg (Bator input file) format
!--------------------------------------------------------------------------------------------------
! Global Variables : 
!--------------------------------------------------------------------------------------------------
! History :
!**************************************************************************************************
subroutine PrintSec3(ParamBator,LPrtSec,LPrtEnt)
 integer(kind=jpim)  :: i
 logical, intent(in) :: ParamBator, LPrtSec, LPrtEnt

if (LPrtSec .or. LPrtEnt) then
  write(*,'("SECTION 3 : ",A)') repeat("-",54)
  write(*,'(TR12,"Length of section 3 (bytes)               = ",I10)') ksec3(1)
  write(*,'(TR12,"Reserved                                  = ",I10)') ksec3(2)
  write(*,'(TR12,"Flag for compressed and/or observed data  = ",I10)') ksec3(4)
  write(*,'(TR12,"Number of data subsets                    = ",I10,/)') ksec3(3)
  if (LprtSec) then
    if (ParamBator) then
      write(*,'("-- List of unexpanded data descriptors for param_bator.cfg",/)')
      do i=1,ktdlen - 1
        write(*,'("codage",1X,I5,2X,I6.6)') i, ktdlst(i)
      enddo
      write(*,'("codage",1X,I5,2X,I6.6,/)') i, ktdlst(i)
    else
      write(*,'("-- List of unexpanded data descriptors :",/)')
      do i=1,ktdlen - 1
        write(*,'(I5,2X,I6.6)') i, ktdlst(i)
      enddo
      write(*,'(I5,2X,I6.6,/)') i, ktdlst(i)
    endif
  endif
  call PrintSec3B(ParamBator,LPrtEnt)
endif
end subroutine PrintSec3


!**************************************************************************************************
! Subroutine/Function/Program | PrintSec3B
! Author                      | F. Guillaume (CNRM/GMAP/OBS)
! Creation Date               | 01/06/2016
!--------------------------------------------------------------------------------------------------
! Description : print second part section 3 of bufr file, expanded sequence descriptor, in simple
!               or in param.cfg (Bator input file) format
!--------------------------------------------------------------------------------------------------
! Global Variables : 
!--------------------------------------------------------------------------------------------------
! History :
!**************************************************************************************************
subroutine PrintSec3B(ParamBator,LPrtSec)
 integer(kind=jpim)  :: i
 logical, intent(in) :: ParamBator, LPrtSec

if (LPrtSec) then
  if (ParamBator) then
    write(*,'("-- List of expanded data descriptors for param_bator.cfg:",/)')
    do i=1,ktdexl - 1
      write(*,'("values",1X,I5,2X,I6.6,2X,A)') i, ktdexp(i), cnames(i)(1:44)
    enddo
    write(*,'("values",1X,I5,2X,I6.6,2X,A,/)') i, ktdexp(i), cnames(i)(1:44)
  else
    write(*,'("-- List of expanded data descriptors :",/)')
    do i=1,ktdexl - 1
      write(*,'(I5,2X,I6.6,2X,A)') i, ktdexp(i), cnames(i)(1:51)
    enddo
    write(*,'(I5,2X,I6.6,2X,A,/)') i, ktdexp(i), cnames(i)(1:51)
  endif
endif
end subroutine PrintSec3B


!**************************************************************************************************
! Subroutine/Function/Program | PrintSec4
! Author                      | F. Guillaume (CNRM/GMAP/OBS)
! Creation Date               | 01/06/2016
!--------------------------------------------------------------------------------------------------
! Description : print section 4 of bufr file, for one or several subsets
!--------------------------------------------------------------------------------------------------
! Global Variables : 
!--------------------------------------------------------------------------------------------------
! History :
!**************************************************************************************************
subroutine PrintSec4(kel,FirstSubset,LastSubset,LPrtSec)
 integer(kind=jpim), intent(in)    :: kel
 logical,intent(in)                :: LPrtSec
 integer(kind=jpim), intent(inout) :: FirstSubset, LastSubset
 integer(kind=jpim)                :: i, j, ichar0, FirstSubsetLoc
 character(len=80)                 :: ccitt

if (LPrtSec) then
  write(*,'("SECTION 4 : ",A)') repeat("-",54)
  if (FirstSubset == 0) then
    FirstSubsetLoc = 1
    LastSubset  = ksup(6)
  else
    FirstSubsetLoc = FirstSubset
  endif
  do j=FirstSubsetLoc, LastSubset
    ccitt = ''
!    if ((LastSubset - FirstSubset) > 1 .and. ksup(6) > 1) write(*,'(/,"Subset ",I4.4)') j
    write(*,'(/,"Subset ",I4.4,"/",I4.4)') j, (LastSubset - FirstSubsetLoc + 1)
    do i=1,ktdexl
      if (cunits(i)(1:4) == 'CCIT') then
        ccitt=adjustl(cvals(NINT(values((j-1)*kel+i))/1000))
        ichar0 = index(ccitt,ACHAR(0))
        if (ichar0 > 0) ccitt(ichar0:80) = REPEAT(" ",(80 - ichar0 + 1))
      endif
      if (values((j-1)*kel+i) == rvind) then
        write(*,'(I6,2X,A20,17X,"MISSING",2X,A)') i,cnames(i), cunits(i)
      else
        if (len(trim(ccitt)) == 0) then
          write(*,'(I6,2X,A20,2X,E22.14E3,2X,A)') i, cnames(i), values((j-1)*kel+i), cunits(i)
        else
          write(*,'(I6,2X,A20,2X,E22.14E3,2X,A,2x,A)') i, cnames(i), values((j-1)*kel+i), cunits(i), trim(ccitt)
          ccitt = ''
        endif
      endif
    enddo
  enddo
  endif
end subroutine PrintSec4


!**************************************************************************************************
! Subroutine/Function/Program | CountBufr
! Author                      | F. Guillaume (CNRM/GMAP/OBS)
! Creation Date               | 01/06/2016
!--------------------------------------------------------------------------------------------------
! Description : check program arguments and eventually adjust number of files and subsets to decode
!--------------------------------------------------------------------------------------------------
! Global Variables : 
!--------------------------------------------------------------------------------------------------
! History :
!**************************************************************************************************
SUBROUTINE CountBufr(nulbufr,FirstFile,LastFile,FirstSubset,LastSubset)
 integer(kind=jpim), intent(in)    :: nulbufr
 integer(kind=jpim), intent(inout) :: FirstFile, LastFile, FirstSubset, LastSubset
 integer(kind=jpim)                :: NbFic, i
 integer(kind=jpim)                :: iword, iret, iret2
 character(len=128)                :: FmtFile, FmtSubset

iret     = 0
iret2    = 0

NbFic    = 0
do while (iret == 0)
  call pbbufr(nulbufr,kbuff,jbyte,kbufl,iret)
  if (iret /= 0) cycle
  NbFic = NbFic + 1
enddo

call pbseek(nulbufr,0,0,iret)

if (FirstFile > 0) then
  if (NbFic < LastFile) then
    FmtFile='("***** ERROR : Asked BUFR files from ",I6.6," to ",I6.6," to expand, but only ",I6," are in InputFile.")'
    write(ErrString,fmt=FmtFile) FirstFile, LastFile, NbFic
    call erreur(trim(ErrString))
  endif
  NbFic    = 0
  do i=1, FirstFile - 1
    call pbbufr(nulbufr,kbuff,jbyte,kbufl,iret)
    NbFic = NbFic + 1
  enddo
else
  FirstFile = 1
  LastFile  = NbFic
endif

do i=FirstFile, LastFile
  call pbbufr(nulbufr,kbuff,jbyte,kbufl,iret)
  iword = kbufl / (jbpw / 8) + 1
  call bus012(iword,kbuff,ksup,ksec0,ksec1,ksec2,iret2)
  if (iret2 /= 0) then
    fmtFile='("***** ERROR  : bus012() - Bufr file n° ",I6.6," is incorrect.")'
    write(ErrString,fmt=FmtFile) i
    call erreur(trim(ErrString))
  else
    if (ksup(6) < LastSubset) then
      FmtSubset='("***** ERROR : Asked subsets from ",I4.4," to ",I4.4," to expand, but only ",I4.4, &
                 & " subsets are in BUFR n° ",I6,".")'
      write(ErrString,fmt=FmtSubset) FirstSubset, LastSubset, ksup(6),i
      call erreur(trim(ErrString))
    endif
  endif
enddo
END SUBROUTINE CountBufr

!**************************************************************************************************
! Subroutine/Function/Program | Erreur
! Author                      | F. Guillaume (CNRM/GMAP/OBS)
! Creation Date               | 01/06/2016
!--------------------------------------------------------------------------------------------------
! Description : Management for runtime error
!--------------------------------------------------------------------------------------------------
! Global Variables : 
!--------------------------------------------------------------------------------------------------
! History :
!**************************************************************************************************
subroutine erreur(message)
  character(len=*), intent(in) :: message

  write (*,'(/,A,/)') message
  print *, "Syntax : DecodBufr -i [path]inputfile [-f n[-m]] [-s n[-m]] [-b] [-h [0][1][2][3][e][4]]"
  print *, "-f is used to specify a particular or a sequence of bufr files (n > 0 & m >= n)"
  print *, "-s is used to specify a particular or a sequence of subsets ( n > 0 & m >= n)"
  print *, "-b is used to print section 3 following the param_bator.cfg input file syntax"
  print *, "-h disable printing of designed sections"
  stop
end subroutine erreur

END MODULE decodtool_module
